#include <fcntl.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <utime.h>

#define PASSWD_FILE "/etc/passwd"
#define SHADOW_FILE "/etc/shadow"

static int check_link_count(const char *file);
static int create_backup(const char *backup, FILE *fp);

int update_passwd(const struct passwd *pw, char *crypt_pw) {
  char file[1024];
  char lock[1024];
  char filename[1024];
  char buf[32];
  char buffer[80];
  char username[32];
  char *pw_rest;
  char *cp;
  int pid;
  int len;
  int retval;
  int has_shadow = 0;
  int fd;
  int mask;
  int continued;
  FILE *fp;
  FILE *out_fp;
  struct stat sb;

  if (access(SHADOW_FILE, F_OK) == 0) {
    has_shadow = 1;
  }
  if (has_shadow) {
    snprintf(filename, sizeof filename, "%s", SHADOW_FILE);
  } else {
    snprintf(filename, sizeof filename, "%s", PASSWD_FILE);
  }
  snprintf(file, sizeof file, "%s.%ld", filename, (long) getpid());
  snprintf(lock, sizeof lock, "%s.lock", filename);
  if ((fd = open(file, O_CREAT|O_EXCL|O_WRONLY, 0600)) == -1) {
    return 0;
  }

  pid = getpid();
  snprintf(buf, sizeof buf, "%d", pid);
  len = strlen(buf) + 1;
  if (write(fd, buf, len) != len) {
    close(fd);
    unlink(file);
    return 0;
  }
  close (fd);
  if (!link(file, lock) == 0) {
    unlink(file);
    return 0;
  }
  if (!check_link_count(file)) {
    unlink(file);
    return 0;
  }
  unlink(file);
// We have a lock, let's make a backup and start updating.
  if (((fp = fopen(filename, "r+")) == 0) || (fstat(fileno(fp), &sb))) {
    unlink(lock);
    return 0;
  }
  snprintf(buf, sizeof buf, "%s-", filename);
  if (create_backup(buf, fp)) {
    fclose(fp);
    unlink(lock);
    return 0;
  }
  snprintf(buf, sizeof buf, "%s+", filename);
  mask = umask(0777);
  out_fp = fopen(buf, "w");
  umask(mask);
  if ((!out_fp) || (fchmod(fileno(out_fp), sb.st_mode & 0777)) || (fchown(fileno(out_fp), sb.st_uid, sb.st_gid))) {
    fclose(fp);
    fclose(out_fp);
    unlink(buf);
    return 0;
  }
  
  continued = 0;
  snprintf(username, sizeof username, "%s:", pw->pw_name);
  rewind(fp);
  while (!feof(fp)) {
    fgets(buffer, sizeof buffer, fp);
    if (!continued) { // Check to see if we're updating this line.
      if (strncmp(username, buffer, strlen(username)) == 0) { // we have a match.
	pw_rest = strchr(buffer, ':');
	*pw_rest++ = '\0';
	pw_rest = strchr(pw_rest, ':');
	fprintf(out_fp, "%s:%s%s", buffer, crypt_pw, pw_rest);
      } else {
        fputs(buffer, out_fp);
      }
    } else {
      fputs(buffer, out_fp);
    }
    if (buffer[strlen(buffer) - 1] == '\n') {
      continued = 0;
    } else {
      continued = 1;
    }
    bzero(buffer, sizeof buffer);
  }
  
  if ((fflush(out_fp)) || (fsync(fileno(out_fp))) || (fclose(out_fp)) || (fclose(fp))) {
    unlink(buf);
    return 0;
  }
  if (rename(buf, filename)) {
    return 0;
  } else {
    unlink(lock);
    return 1;
  }
}

static int check_link_count(const char *file) {
  struct stat sb;
  if (stat(file, &sb) != 0)
    return 0;
  if (sb.st_nlink != 2)
    return 0;
  return 1;
}                                                                         

static int create_backup(const char *backup, FILE *fp) {
  struct stat sb;
  struct utimbuf ub;
  FILE *bkfp;
  int c, mask;

  if (fstat(fileno(fp), &sb))
    return -1;

  mask = umask(077);
  bkfp = fopen(backup, "w");
  umask(mask);
  if (!bkfp)
    return -1;

  /* TODO: faster copy, not one-char-at-a-time.  --marekm */
  rewind(fp);
  while ((c = getc(fp)) != EOF) {
    if (putc(c, bkfp) == EOF)
      break;
  }
  if (c != EOF || fflush(bkfp)) {
    fclose(bkfp);
    return -1;
  }
  if (fclose(bkfp))
    return -1;

  ub.actime = sb.st_atime;
  ub.modtime = sb.st_mtime;
  utime(backup, &ub);
  return 0;
}
